FRAMES_PATH_COL_NAME = "frames_path"
MULTI_FRAME_LABELS_EXT = ".multi-frame-labels.npy"
MULTI_FRAME_LABELS_PATH_COL_NAME = "multi_frame_labels_path"
ANNOTATION_CSV_FILENAME = "y.csv"
SAMPLE_IDS_ARRAY_FILENAME = "sample_ids.npy"
INDS_IN_SAMPLE_ARRAY_FILENAME = "inds_in_sample.npy"
WINDOW_INDS_ARRAY_FILENAME = "window_inds.npy"
FRAME_CLASSIFICATION_DATASET_AUDIO_FORMAT = "wav"
